#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>
#include <math.h>

#include "../../netflixcore.h"
#include "svd.h"

int svdhelper(rlist *rtmp, int K, double lam, double bias) {
  int i, j, k, k2, M_known, itmp;
  feature *Q_known, *A_known, *KK;
  svdinfo *svdtmp1, *svdtmp2;
  double tmp;

  svdtmp1 = (svdinfo*)(rtmp->additional);
  M_known = svdtmp1->nvisible;
  Q_known = malloc(M_known*K*sizeof(feature));
  A_known = malloc(1*M_known*sizeof(feature));
  KK = malloc(K*K*sizeof(feature));

  itmp = 0;
  for(i = 0; i < rtmp->length; i++) {
    if(rtmp->ratings[i].visible) {
      svdtmp2 = svd_infoformovie(rtmp->ratings[i].id);
      for(k = 0; k < K; k++) {
	Q_known[itmp+k*M_known] = svdtmp2->features[k];
      }
      A_known[itmp] = rtmp->ratings[i].stars - bias
	- svdtmp1->bias - svdtmp2->bias;
      itmp++;
    }
  }
    
  for(k = 0; k < K; k++) {
    for(k2 = 0; k2 < K; k2++) {
      KK[k+k2*K] = 0.0;
      for(i = 0; i < M_known; i++) {
	KK[k+k2*K] += Q_known[i+k*M_known]*Q_known[i+k2*M_known];
      }
      if(k == k2) {
	KK[k+k2*K] += lam*M_known;
      }
    }
  }
    
  if(invmat(KK,K,K) != 0) {
    perror("Singular matrix :-(\n");
    free(KK);
    free(Q_known); free(A_known);
    return -1;
  }
    
  for(k = 0; k < K; k++) {
    svdtmp1->features[k] = 0.0;
    for(i = 0; i < M_known; i++) {
      tmp = 0.0;
      for(k2 = 0; k2 < K; k2++) {
	tmp += KK[k+k2*K]*Q_known[i+k2*M_known];
      }
      svdtmp1->features[k] += tmp*A_known[i];
    }
  }

  free(KK); free(Q_known); free(A_known);
  return 0;
}

/* Loads a hypothetical person (and their ratings) from a file.
   Leaves space for up to 1000 ratings */
rlist* svd_loaduser(FILE *f, double lam, double bias) {
  rlist *rtmp;
  svdinfo *svdtmp, *svdtmp2;
  int mid, stars, length, ret, i, K;
  double lam3, tmp;

  /* Probably shouldn't hardcode all of these... */
  lam3 = 10;
  K = NFEATURES;

  /* Check the file is open */
  if(f == NULL) {
    perror("File problems with ratings");
    return NULL;
  }

  /* How many movies did he rate? */
  ret = fscanf(f, "%d ratings", &length);
  assert(ret != EOF);

  /* Make an rlist and svdinfo struct for this person */
  rtmp = malloc(sizeof(rlist));
  rtmp->length = length;
  rtmp->ratings = malloc(sizeof(rating)*length);
  rtmp->additional = malloc(sizeof(svdinfo));
  svdtmp = (svdinfo*)rtmp->additional;
  svdtmp->bias = 0.0;
  svdtmp->err = 0.0;
  svdtmp->nvisible = length;
  svdtmp->features = malloc(sizeof(feature)*NFEATURES);
  /* Should probably check that the mallocs worked... */

  /* Read in the ratings, initialize the SVD stuff */
  for(i = 0; i < length; i++) {
    ret = fscanf(f, "%d %d", &mid, &stars);
    assert(ret != EOF);
    mid--;
    rtmp->ratings[i].stars = stars;
    rtmp->ratings[i].id = mid;
    rtmp->ratings[i].visible = 1;
    svdtmp2 = (svdinfo*)(nf_ratingsformovie(rtmp->ratings[i].id)->additional);
    svdtmp->bias += rtmp->ratings[i].stars - bias - svdtmp2->bias;
  }
  svdtmp->bias /= lam3 + svdtmp->nvisible;
  for(i = 0; i < NFEATURES; i++)
    svdtmp->features[i] = rand_0_1();
  /* Make sure that's all there was in the file */
  while((i = fgetc(f)) != EOF)
    assert(isspace(i));

  /* Solve for the feature vector using ALS */
  if(svdhelper(rtmp, K, lam, bias) != 0) {
    fprintf(stderr, "Problem solving for SVD vector\n");
  }

  fclose(f);
  return rtmp;
}

TopNNode* svd_recommend(rlist *rtmp, int n, double bias) {
  int i, k, alreadyrated;
  double tmp;
  svdinfo *svdtmp, *svdtmp2;
  TopNNode *toplist = NULL;

  svdtmp = (svdinfo*)(rtmp->additional);
 
  /* Make a list of top ratings (predicted) */
  for(i = 0; i < NMOVIES; i++) {
    svdtmp2 = svd_infoformovie(i);
    tmp = 0.0;
    for(k = 0; k < NFEATURES; k++) {
      tmp += svdtmp->features[k]*svdtmp2->features[k];
    }
    tmp += bias + svdtmp->bias + svdtmp2->bias;
    alreadyrated = 0;
    for(k = 0; k < rtmp->length; k++) {
      if(i == rtmp->ratings[k].id) {
        alreadyrated = 1;
        break;
        /* printf("%d %f\n", rtmp->ratings[k].stars, tmp); */
      }
    }
    if(!alreadyrated) {
      toplist = topn_addrating(toplist, tmp, i, n);
    }
  }

  return toplist;
}

/*
 * Shows the top n movies along the dim dimension
 * (bottom n movies if n < 0, all dimensions if dim == -1)
 */
int svd_dimensions(int dim, int n) {
  TopNNode *toplist, *p;
  int i, LIM, flip, j, *mids;
  svdinfo *svdtmp;
  char **titles;
  char *header;

  if(dim < -1 || dim >= NFEATURES) {
    fprintf(stderr, "Invalid dimension!\n");
    return 1;
  }

  if(dim == -1) {
    LIM = NFEATURES;
    i = 0;
  }
  else {
    i = dim;
    LIM = i+1;
  }

  flip = 1;
  header = "Top";
  if(n < 0) {
    header = "Bottom";
    flip = -1;
    n *= -1;
  }

  titles = nf_alloctitles(n);

  /* For all the dimensions in question */
  for(; i < LIM; i++) {
    printf("\n%s %d movies along the %d dimension:\n", header, n, i);
    toplist = NULL;
    for(j = 0; j < NMOVIES; j++) {
      svdtmp = svd_infoformovie(j);
      toplist = topn_addrating(toplist, svdtmp->features[i]*flip, j, n);
    }
    mids = topn_findoffsets(toplist, n);
    nf_findtitles(mids, titles, n);
    for(p = toplist; p != NULL; p = p->next) {
      printf("%.4f\t%d\t%s\n", p->rating*flip, p->mid, titles[p->offset]);
    }
    topn_freeall(toplist, n);
  }

  nf_freetitles(titles, n); free(mids);
  return 0;
}

