/*
 * Implements a "Top-N" data structure with a linked list, so that we can 
 * figure out what your top 5 (or 10 or w/e) picks for movies would be.
 */

#include "svd.h"
#include <stdlib.h>

/* Inserts a rating into the list, making sure the list never exceeds n items */
TopNNode* topn_addrating(TopNNode *first, double rating, int mid, int n) {
  TopNNode *p, *p2;

  /* Insert at beginning? */
  if(first == NULL || first->rating < rating) {
    /*printf("mallocing %d\n", mid);*/
    p = malloc(sizeof(TopNNode));
    p->rating = rating;
    p->mid = mid;
    p->next = first;
    topn_trim(p, n);
    return p;
  }

  for(p = first; p->next != NULL && p->next->rating >= rating; p = p->next) {}
  /*printf("mallocing %d\n", mid);*/
  p2 = malloc(sizeof(TopNNode));
  p2->rating = rating;
  p2->mid = mid;
  p2->next = p->next;
  p->next = p2;
  topn_trim(first, n);
  return first;
}

/* Trims off the last item (if there are more than n) */
void topn_trim(TopNNode *first, int n) {
  TopNNode *p;
  int i;

  for(i = 0, p = first; p->next != NULL && p->next->next != NULL; i++, p = p->next);
  if(p->next == NULL) return;
  if(i+2 <= n) return;
  /*printf("freeing %d\n", p->next->mid);*/
  free(p->next);
  p->next = NULL;
}

/* Frees all the ratings in the list (up to n of them) */
void topn_freeall(TopNNode *first, int n) {
  TopNNode *p;
  int i, j;

  for(i = n-1; i >= 0; i--) {
    for(j = 0, p = first; j < i && p != NULL; j++, p = p->next) {
      /*printf("%d\n", p->mid);*/
    }
    if(p != NULL) {
      /*printf("freeing %d\n", p->mid);*/
      free(p);
    }
  }
}

static int cmpnodes(const void *a, const void *b) {
  TopNNode **t1, **t2;

  t1 = (TopNNode**)a;
  t2 = (TopNNode**)b;

  if((*t1)->mid < (*t2)->mid) return -1;
  if((*t1)->mid > (*t2)->mid) return 1;
  return 0;
}

int* topn_findoffsets(TopNNode *first, int n) {
  TopNNode **tmp, *p;
  int i, *mids;

  tmp = malloc(sizeof(TopNNode*)*n);
  mids = malloc(sizeof(int)*n);
  for(i = 0, p = first; i < n && p != NULL; i++, p = p->next) {
    tmp[i] = p;
  }
  qsort(tmp, n, sizeof(TopNNode*), cmpnodes);
  for(i = 0; i < n; i++) {
    tmp[i]->offset = i;
    mids[i] = tmp[i]->mid;
  }
  free(tmp);
  return mids;
}
