#include "flagv.h"
#include <stdlib.h>
#include <string.h>
#include <assert.h>

/* Create a flagvec with a given capacity (initialized to all 0's) */
flagvec flagv_create(int capacity) {
  flagvec fv;
  int nbytes = capacity / 8; /* We can fit 8 flags per byte */
  if(capacity % 8 > 0) nbytes++;
  fv = malloc(nbytes*sizeof(char));
  if(fv != NULL) memset(fv, 0, sizeof(char)*nbytes);
  return fv;
}

/* Find the value of the flag at index i in flagvec fv */
inline int flagv_get(flagvec fv, int i) {
  int byteoffset = i / 8;
  int bitoffset = i % 8;
  return (fv[byteoffset] & (1 << bitoffset)) >> bitoffset;
}

/* Set the value of the flag at index i to be f */
inline void flagv_set(flagvec fv, int i, int f) {
  int byteoffset = i / 8;
  int bitoffset = i % 8;
  assert(f == 0 || f == 1);
  fv[byteoffset] &= ~(0 << bitoffset);
  fv[byteoffset] |= f << bitoffset;
}
