/* topn.h - Peter Nelson
 *
 * Headers for constructing arbitrarily large "Top-10" style lists of movies
 *
 */
#ifndef TOPN_H
#define TOPN_H

/* Top-n Linked List Node */
struct TopNNode {
  double rating;         /* Rating (stars) */
  int mid;               /* Movie ID */
  int offset;            /* Used for finding names */
  struct TopNNode *next;
};
typedef struct TopNNode TopNNode;

/* Add a movie to the list */
TopNNode* topn_addmovie(TopNNode *first, double rating, int mid, int n);

/* Free all the memory in the list (of size n) */
void topn_freeall(TopNNode *first, int n);

/* Trim the last item from the list, if the list is too long */
void topn_trim(TopNNode *first, int n);

/* Finds the number of items in the list (which might be less than n) */
int topn_len(TopNNode *first);

/* Used so that you can find the titles of the movies */
int* topn_findoffsets(TopNNode *first, int n);

#endif
